__all__ = [
    "__title__", "__summary__", "__uri__", "__author__", "__email__", "__version__",
]

__title__ = "covid-shared"
__summary__ = "Shared utilities for the covid model pipeline."
__uri__ = "https://stash.ihme.washington.edu/projects/CVD19/repos/covid-shared/browse"

__version__ = '1.0.50'

__author__ = "James Collins"
__email__ = "collijk@uw.edu"
